r"""Test parse_key."""

from pyrime.parse_key import parse_key


class Test:
    r"""Test."""

    @staticmethod
    def test_parse_key() -> None:
        r"""Test parse key.

        :rtype: None
        """
        assert parse_key("c-^", set()) == parse_key("6", {"Control"})



"""Ansi
=======

Convert ANSI escape code to prompt-toolkit class.

Experimental!

`<https://github.com/prompt-toolkit/ptpython/issues/525>`_
"""

import re
from functools import reduce

from colorama import Back, Fore, Style
from prompt_toolkit.formatted_text import AnyFormattedText

CODES = reduce(
    lambda x, y: x | y,
    map(
        lambda x: dict(
            (
                v,
                x[0]
                + k.lower()
                .replace("bright", "bold")
                .replace("dim", "italic")
                .replace(
                    "normal",
                    "nobold noitalic nounderline noblink noreverse nohidden"
                    " fg:white bg:black",
                )
                .replace("reset_all", "")
                .replace(
                    "reset", "black" if x[0].startswith("bg") else "white"
                )
                .replace("light", "bright")
                .replace("_ex", ""),
            )
            for k, v in vars(x[1]).items()
        ),
        {"fg:ansi": Fore, "bg:ansi": Back, "": Style}.items(),
    ),
)
PAT = re.compile("(" + "|".join(CODES).replace("[", "\\[") + ")")


def ansi2prompt_toolkit(
    ansi: str, prefix: str = ""
) -> tuple[AnyFormattedText, dict[str, str]]:
    """Ansi2prompt toolkit.

    :param ansi:
    :type ansi: str
    :param prefix:
    :type prefix: str
    :rtype: tuple[AnyFormattedText, dict[str, str]]
    """
    formats = []
    styles = {}
    num = 0
    class_name = prefix + str(num)
    text = ""
    style_name = ""
    for token in filter(len, PAT.split(ansi)):
        code_name = CODES.get(token)
        if code_name is None:
            text = token.replace("\n>>> ", "")
            formats += [("class:" + class_name, text)]
            styles |= {class_name: style_name}
            class_name += "." + str(num)
            text = ""
        if code_name == "":
            num += 1
            class_name = prefix + str(num)
            text = ""
        if code_name:
            style_name = code_name
            formats += [("class:" + class_name, text)]
            styles |= {class_name: style_name}
            class_name += "." + str(num)
            text = ""
    return formats, styles



r"""Autopair
============

Refer `zsh-autopair <https://github.com/hlissner/zsh-autopair>`_.
"""

from prompt_toolkit.key_binding.bindings.named_commands import (
    backward_char,
    backward_delete_char,
    delete_char,
    forward_char,
)
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from .. import Rime
from ..utils.condition import InsertMode


def autopair(rime: Rime) -> None:
    r"""Autopair.

    :param rime:
    :type rime: Rime
    :rtype: None
    """
    repl = rime.repl

    @repl.add_key_binding("c-h", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        cr: str = b.document.current_char
        if b.document.cursor_position_col == 0:
            cl: str = ""
        else:
            cl: str = b.document.char_before_cursor
        if b.document.cursor_position_col <= 1:
            cl2: str = ""
        else:
            backward_char(event)
            cl2: str = b.document.char_before_cursor
            forward_char(event)
        for c0, c1 in ["[]", "()", "{}", "''", "``", '""']:
            if cl == c0 and cr == c1:
                delete_char(event)
                break
            elif cl == c1 and cl2 == c0:
                backward_delete_char(event)
                break
        backward_delete_char(event)

    @repl.add_key_binding("[", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.cli.current_buffer.insert_text("[]")
        backward_char(event)

    @repl.add_key_binding("]", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != "]":
            b.insert_text("]")
        else:
            forward_char(event)

    @repl.add_key_binding("(", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.cli.current_buffer.insert_text("()")
        backward_char(event)

    @repl.add_key_binding(")", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != ")":
            b.insert_text(")")
        else:
            forward_char(event)

    @repl.add_key_binding("{", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.cli.current_buffer.insert_text("{}")
        backward_char(event)

    @repl.add_key_binding("}", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != "}":
            b.insert_text("}")
        else:
            forward_char(event)

    @repl.add_key_binding("'", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != "'":
            b.insert_text("''")
            backward_char(event)
        else:
            forward_char(event)

    @repl.add_key_binding("`", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != "`":
            b.insert_text("``")
            backward_char(event)
        else:
            forward_char(event)

    @repl.add_key_binding('"', filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char != '"':
            b.insert_text('""')
            backward_char(event)
        else:
            forward_char(event)



r"""Viemacs
===========

Use ``EmacsInsertMode()`` to replace ``ViInsertMode()``

Refer `vim-rsi <https://github.com/tpope/vim-rsi>`_.
"""

from prompt_toolkit.clipboard import ClipboardData
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import (
    EmacsInsertMode,
    ViNavigationMode,
    in_paste_mode,
)
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.selection import SelectionType

from .. import Rime


def viemacs(rime: Rime) -> None:
    repl = rime.repl

    @repl.add_key_binding("escape", filter=EmacsInsertMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.VI
        event.app.vi_state.input_mode = InputMode.NAVIGATION
        rime.conditional_disable()

    @repl.add_key_binding("i", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        rime.conditional_enable()

    @repl.add_key_binding("a", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        event.current_buffer.cursor_position += (
            event.current_buffer.document.get_cursor_right_position()
        )
        rime.conditional_enable()

    @repl.add_key_binding("I", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        event.current_buffer.cursor_position += (
            event.current_buffer.document.get_start_of_line_position(
                after_whitespace=True
            )
        )
        rime.conditional_enable()

    @repl.add_key_binding("A", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        event.current_buffer.cursor_position += (
            event.current_buffer.document.get_end_of_line_position()
        )
        rime.conditional_enable()

    @repl.add_key_binding("o", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        event.current_buffer.insert_line_below(copy_margin=not in_paste_mode())
        rime.conditional_enable()

    @repl.add_key_binding("O", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        event.current_buffer.insert_line_above(copy_margin=not in_paste_mode())
        rime.conditional_enable()

    @repl.add_key_binding("s", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        text = event.current_buffer.delete(count=event.arg)
        event.app.clipboard.set_text(text)
        rime.conditional_enable()

    @repl.add_key_binding("C", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        buffer = event.current_buffer

        deleted = buffer.delete(
            count=buffer.document.get_end_of_line_position()
        )
        event.app.clipboard.set_text(deleted)
        rime.conditional_enable()

    @repl.add_key_binding("c", "c", filter=ViNavigationMode())  # type: ignore
    @repl.add_key_binding("S", filter=ViNavigationMode())  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        event.app.editing_mode = EditingMode.EMACS
        event.app.vi_state.input_mode = InputMode.INSERT
        buffer = event.current_buffer
        # We copy the whole line.
        data = ClipboardData(buffer.document.current_line, SelectionType.LINES)
        event.app.clipboard.set_data(data)

        # But we delete after the whitespace
        buffer.cursor_position += buffer.document.get_start_of_line_position(
            after_whitespace=True
        )
        buffer.delete(count=buffer.document.get_end_of_line_position())
        rime.conditional_enable()



r"""Draw UI
===========
"""

from dataclasses import dataclass

from wcwidth import wcswidth

from . import Context


@dataclass
class UI:
    r"""UI."""

    indices: list[str] = None  # type: ignore
    left: str = "<|"
    right: str = "|>"
    left_sep: str = "["
    right_sep: str = "]"
    cursor: str = "|"

    def __post_init__(self) -> None:
        if self.indices is None:
            self.indices = ["①", "②", "③", "④", "⑤", "⑥", "⑦", "⑧", "⑨", "⓪"]


def draw_ui(context: Context, ui: UI) -> tuple[list[str], int]:
    r"""Draw UI.

    :param context:
    :type context: Context
    :param ui:
    :type ui: UI
    :rtype: tuple[list[str], int]
    """
    preedit = context.composition.preedit
    preedit = (
        preedit[0 : context.composition.cursor_pos]
        + ui.cursor
        + preedit[context.composition.cursor_pos :]
    )
    candidates = context.menu.candidates
    candidates_ = ""
    indices = ui.indices
    for index, candidate in enumerate(candidates):
        text = indices[index] + " " + candidate.text
        if candidate.comment:
            text = text + " " + candidate.comment
        if context.menu.highlighted_candidate_index == index:
            text = ui.left_sep + text
        elif context.menu.highlighted_candidate_index + 1 == index:
            text = ui.right_sep + text
        else:
            text = " " + text
        candidates_ = candidates_ + text
    if (
        context.menu.num_candidates
        == context.menu.highlighted_candidate_index + 1
    ):
        candidates_ = candidates_ + ui.right_sep
    else:
        candidates_ = candidates_ + " "
    col = 0
    left = ui.left
    if context.menu.page_no != 0:
        num = wcswidth(ui.left)
        candidates_ = left + candidates_
        preedit = " " * num + preedit
        col = col - num
    if not context.menu.is_last_page and context.menu.num_candidates > 0:
        candidates_ = candidates_ + ui.right
    return [preedit, candidates_], col



r"""Smartinput
==============

Add spaces around operators.

Refer `vim-smartinput <https://github.com/kana/vim-smartinput>`_.
"""

from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from .. import Rime
from ..utils.condition import InsertMode


def smartinput(rime: Rime) -> None:
    r"""Smartinput.

    :param rime:
    :type rime: Rime
    :rtype: None
    """
    repl = rime.repl

    # One {{{1 #
    @repl.add_key_binding(",", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char == " ":
            b.insert_text(",")
        else:
            b.insert_text(", ")

    # 1}}} One #

    # Two {{{1 #
    # Operation {{{2 #
    @repl.add_key_binding("+", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.current_char == " ":
            b.insert_text("+")
        else:
            b.insert_text(" + ")

    @repl.add_key_binding("@", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if (
            b.document.char_before_cursor == " "
            or b.document.cursor_position_col == 0
        ):
            b.insert_text("@")
        else:
            b.insert_text(" @ ")

    @repl.add_key_binding("*", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("*")
        else:
            b.insert_text(" * ")

    @repl.add_key_binding("*", "*", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("**")
        else:
            b.insert_text(" ** ")

    @repl.add_key_binding("/", "/", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("//")
        else:
            b.insert_text(" // ")

    @repl.add_key_binding("%", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if (
            b.document.char_before_cursor == " "
            or b.document.cursor_position_col == 0
        ):
            b.insert_text("%")
        else:
            b.insert_text(" % ")

    @repl.add_key_binding("&", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("&")
        else:
            b.insert_text(" & ")

    @repl.add_key_binding("|", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("|")
        else:
            b.insert_text(" | ")

    @repl.add_key_binding("^", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("^")
        else:
            b.insert_text(" ^ ")

    @repl.add_key_binding("<", "<", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("<<")
        else:
            b.insert_text(" << ")

    @repl.add_key_binding(">", ">", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text(">>")
        else:
            b.insert_text(" >> ")

    # 1}}} Operation #

    # Relation {{{2 #
    @repl.add_key_binding("<", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("<")
        else:
            b.insert_text(" < ")

    @repl.add_key_binding(">", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text(">")
        else:
            b.insert_text(" > ")

    @repl.add_key_binding(":", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text(":=")
        else:
            b.insert_text(" := ")

    @repl.add_key_binding("=", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("==")
        else:
            b.insert_text(" == ")

    @repl.add_key_binding("!", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("!=")
        else:
            b.insert_text(" != ")

    @repl.add_key_binding("<", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("<=")
        else:
            b.insert_text(" <= ")

    @repl.add_key_binding(">", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text(">=")
        else:
            b.insert_text(" >= ")

    # 1}}} Relation #

    # Assign {{{2 #
    @repl.add_key_binding("=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("=")
        else:
            b.insert_text(" = ")

    @repl.add_key_binding("+", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("+=")
        else:
            b.insert_text(" += ")

    @repl.add_key_binding("-", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("-=")
        else:
            b.insert_text(" -= ")

    @repl.add_key_binding("@", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("@=")
        else:
            b.insert_text(" @= ")

    @repl.add_key_binding("*", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("*=")
        else:
            b.insert_text(" *= ")

    @repl.add_key_binding("/", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("/=")
        else:
            b.insert_text(" /= ")

    @repl.add_key_binding("*", "*", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("**=")
        else:
            b.insert_text(" **= ")

    @repl.add_key_binding("/", "/", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("//=")
        else:
            b.insert_text(" //= ")

    @repl.add_key_binding("%", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("%=")
        else:
            b.insert_text(" %= ")

    @repl.add_key_binding("&", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("&=")
        else:
            b.insert_text(" &= ")

    @repl.add_key_binding("|", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("|=")
        else:
            b.insert_text(" |= ")

    @repl.add_key_binding("^", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("^=")
        else:
            b.insert_text(" ^= ")

    @repl.add_key_binding("<", "<", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text("<<=")
        else:
            b.insert_text(" <<= ")

    @repl.add_key_binding(">", ">", "=", filter=rime.filter(InsertMode))  # type: ignore
    def _(event: KeyPressEvent) -> None:
        """.

        :param event:
        :type event: KeyPressEvent
        :rtype: None
        """
        b = event.cli.current_buffer
        if b.document.char_before_cursor == " ":
            b.insert_text(">>=")
        else:
            b.insert_text(" >>= ")

    # 2}}} Assign #
    # 1}}} Two #


# ex: foldmethod=marker



r"""Provide ``__version__`` for
`importlib.metadata.version() <https://docs.python.org/3/library/importlib.metadata.html#distribution-versions>`_.
"""

from dataclasses import dataclass

@dataclass
class Composition:
    length: int
    cursor_pos: int
    sel_start: int
    sel_end: int
    preedit: str

@dataclass
class Candidate:
    text: str
    comment: str | None

@dataclass
class Menu:
    page_size: int
    page_no: int
    is_last_page: int
    highlighted_candidate_index: int
    num_candidates: int
    select_keys: str | None
    candidates: list[Candidate]

@dataclass
class Context:
    composition: Composition
    menu: Menu

@dataclass
class Commit:
    text: str

def init(
    shared_data_dir: str,
    user_data_dir: str,
    log_dir: str,
    distribution_name: str,
    distribution_code_name: str,
    distribution_version: str,
    app_name: str,
    min_log_level: int,
) -> None: ...
def createSession() -> int: ...
def destroySession(session_id: int) -> None: ...
def getCurrentSchema(session_id: int) -> str: ...
def getSchemaList() -> list[tuple[str, str]]: ...
def selectSchema(session_id: int, name: str) -> None: ...
def processKey(session_id: int, keycode: int, mask: int) -> None: ...
def getContext(session_id: int) -> Context: ...
def getCommit(session_id: int) -> Commit: ...
def commitComposition(session_id: int) -> bool: ...
def clearComposition(session_id: int) -> None: ...



r"""This module can be called by
`python -m <https://docs.python.org/3/library/__main__.html>`_.
"""

import os
from dataclasses import dataclass

from platformdirs import user_data_path

shared_data_dir = ""
eprefix = os.getenv(
    "PREFIX",
    os.path.dirname(os.path.dirname(os.getenv("SHELL", "/bin/sh"))),
)
for prefix in [
    # /usr merge: /usr/bin/sh -> /usr/share/rime-data
    os.path.join(eprefix, "share"),
    # non /usr merge: /bin/sh -> /usr/share/rime-data
    os.path.join(eprefix, "usr/share"),
    "/sdcard",
]:
    path = os.path.expanduser(os.path.join(prefix, "rime-data"))
    if os.path.isdir(path):
        shared_data_dir = path
        break

user_data_dir = ""
for prefix in [
    "~/.config/ibus",
    "~/.local/share/fcitx5",
    "~/.config/fcitx",
    "/sdcard",
]:
    path = os.path.expanduser(os.path.join(prefix, "rime"))
    if os.path.isdir(path):
        user_data_dir = path
        break


@dataclass
class Traits:
    shared_data_dir: str = shared_data_dir
    user_data_dir: str = user_data_dir
    log_dir: str = str(user_data_path("ptpython") / "rime")
    distribution_name: str = "Rime"
    distribution_code_name: str = "pyrime"
    distribution_version: str = "0.0.1"
    app_name: str = "rime.pyrime"
    min_log_level: int = 3


if __name__ == "__main__":
    print(Traits())



r"""Filter
==========
"""

from prompt_toolkit.filters import Condition, EmacsInsertMode, ViInsertMode


def any_condition(*conditions: Condition) -> Condition:
    r"""Any condition.

    :param conditions:
    :type conditions: Condition
    :rtype: Condition
    """

    @Condition
    def _(conditions: tuple[Condition, ...] = conditions) -> bool:
        r""".

        :param conditions:
        :type conditions: tuple[Condition, ...]
        :rtype: bool
        """
        return any(condition() for condition in conditions)

    return _


def all_condition(*conditions: Condition) -> Condition:
    r"""All condition.

    :param conditions:
    :type conditions: Condition
    :rtype: Condition
    """

    @Condition
    def _(conditions: tuple[Condition, ...] = conditions) -> bool:
        r""".

        :param conditions:
        :type conditions: tuple[Condition, ...]
        :rtype: bool
        """
        return all(condition() for condition in conditions)

    return _


InsertMode = any_condition(EmacsInsertMode(), ViInsertMode())



r"""Autosuggestion
==================

Refer `zsh-autosuggestions <https://github.com/zsh-users/zsh-autosuggestions>`_.
"""

import re

from prompt_toolkit.filters import (
    Condition,
    ViInsertMode,
    emacs_mode,
)
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.keys import Keys
from ptpython.repl import PythonRepl


def autosuggestion(repl: PythonRepl) -> None:
    r"""Autosuggestion.

    :param repl:
    :type repl: PythonRepl
    :rtype: None
    """

    @Condition
    def suggestion_available() -> bool:
        """Suggestion available.

        :rtype: bool
        """
        app = repl.app
        return (
            app.current_buffer.suggestion is not None
            and len(app.current_buffer.suggestion.text) > 0
            and app.current_buffer.document.is_cursor_at_the_end
        )

    @repl.add_key_binding(
        "right",
        filter=suggestion_available & ViInsertMode(),  # type: ignore
    )
    @repl.add_key_binding("right", filter=suggestion_available & emacs_mode)  # type: ignore
    @repl.add_key_binding("c-f", filter=suggestion_available & ViInsertMode())  # type: ignore
    @repl.add_key_binding("c-f", filter=suggestion_available & emacs_mode)  # type: ignore
    def _(event: "KeyPressEvent") -> None:
        """.

        :param event:
        :type event: "KeyPressEvent"
        :rtype: None
        """
        b = event.current_buffer
        suggestion = b.suggestion

        if suggestion and event.arg > 0:
            b.insert_text(
                suggestion.text[0 : min(event.arg, len(suggestion.text))]
            )

    @repl.add_key_binding(
        "c-]",
        Keys.Any,  # type: ignore
        filter=suggestion_available & emacs_mode,  # type: ignore
    )
    def _(event: "KeyPressEvent") -> None:
        """.

        :param event:
        :type event: "KeyPressEvent"
        :rtype: None
        """
        b = event.current_buffer
        suggestion = b.suggestion

        # don't support event.arg
        if suggestion and event.arg > 0:
            t = re.split(event.data, suggestion.text)
            b.insert_text(next(x for x in t if x))
            if len(t) != 1:
                b.insert_text(event.data)



r"""Test draw UI."""

from pyrime import Candidate, Composition, Context, Menu
from pyrime.draw_ui import UI, draw_ui


class Test:
    r"""Test."""

    @staticmethod
    def test_draw_ui() -> None:
        r"""Test draw UI.

        :rtype: None
        """
        context = Context(
            composition=Composition(
                length=1,
                cursor_pos=1,
                sel_start=0,
                sel_end=1,
                preedit="w",
            ),
            menu=Menu(
                page_size=10,
                page_no=0,
                is_last_page=0,
                highlighted_candidate_index=0,
                num_candidates=10,
                select_keys=None,
                candidates=[
                    Candidate(text="我", comment=None),
                    Candidate(text="为", comment=None),
                    Candidate(text="玩", comment=None),
                    Candidate(text="问", comment=None),
                    Candidate(text="无", comment=None),
                    Candidate(text="万", comment=None),
                    Candidate(text="完", comment=None),
                    Candidate(text="网", comment=None),
                    Candidate(text="王", comment=None),
                    Candidate(text="外", comment=None),
                ],
            ),
        )
        ui = UI()
        lines = ["w|", "[① 我]② 为 ③ 玩 ④ 问 ⑤ 无 ⑥ 万 ⑦ 完 ⑧ 网 ⑨ 王 ⓪ 外 |>"]
        assert draw_ui(context, ui) == (lines, 0)



r"""Parse Key
=============
"""

import json
import os

json_dir = os.path.join(os.path.dirname(__file__), "assets", "json")
with open(os.path.join(json_dir, "keys.json")) as f:
    keys: dict = json.load(f)
with open(os.path.join(json_dir, "modifiers.json")) as f:
    modifiers_: list = json.load(f)  # type: ignore

SHIFT_CR = ["escape", *"[13;2u"]
ALT_CR = ["escape", "enter"]
ALT_SHIFT_CR = ["escape", *"[13;4u"]
CONTROL_CR = ["escape", *"[13;5u"]
CONTROL_SHIFT_CR = ["escape", *"[13;6u"]
CONTROL_ALT_CR = ["escape", *"[13;7u"]
CONTROL_ALT_SHIFT_CR = ["escape", *"[13;8u"]


def parse_key(key: str, modifiers: set[str]) -> tuple[int, int]:
    r"""Parse key. Convert prompt-toolkit key name to rime key code and mask.

    Refer `<https://python-prompt-toolkit.readthedocs.io/en/master/pages/advanced_topics/key_bindings.html#list-of-special-keys>`_

    :param key:
    :type key: str
    :param modifiers:
    :type modifiers: set[str]
    :rtype: tuple[int, int]
    """
    key = {"c-i": "tab", "c-m": "enter", "c-@": "c-space"}.get(key, key)
    if key.startswith("c-"):
        modifiers |= {"Control"}
        _, _, key = key.partition("c-")
        key = {"space": " ", "^": "6", "-": "_"}.get(key, key)
    if key.startswith("s-"):
        modifiers |= {"Shift"}
        _, _, key = key.partition("s-")
    if len(key) > 1:
        key = key.capitalize()
    key = {
        "Enter": "Return",
        "Pageup": "Page_Up",
        "Pagedown": "Page_Down",
    }.get(key, key)
    keycode = keys.get(key, ord(key[0]))
    mask = 0
    for modifier in modifiers:
        index = modifiers_.index(modifier)
        if index != -1:
            mask += 2**index
    return keycode, mask


def parse_keys(keys: list[str]) -> tuple[int, int]:
    r"""Parse keys.

    :param keys:
    :type keys: list[str]
    :rtype: tuple[int, int]
    """
    if keys == SHIFT_CR:
        key = "enter"
        modifiers = {"Shift"}
    elif keys == ALT_SHIFT_CR:
        key = "enter"
        modifiers = {"Shift", "Alt"}
    elif keys == CONTROL_CR:
        key = "enter"
        modifiers = {"Control"}
    elif keys == CONTROL_SHIFT_CR:
        key = "enter"
        modifiers = {"Shift", "Control"}
    elif keys == CONTROL_ALT_CR:
        key = "enter"
        modifiers = {"Alt", "Control"}
    elif keys == CONTROL_ALT_SHIFT_CR:
        key = "enter"
        modifiers = {"Shift", "Alt", "Control"}
    elif len(keys) == 2 and keys[0] == "escape":
        key = keys[1]
        modifiers = {"Alt"}
    elif len(keys) == 1:
        key = keys[0]
        modifiers = set()
    else:
        raise NotImplementedError
    return parse_key(key, modifiers)



"""Insert
=========
"""

from prompt_toolkit.key_binding.key_processor import KeyPressEvent


def insert(event: KeyPressEvent, pre: str, post: str) -> None:
    """Insert.

    :param event:
    :type event: KeyPressEvent
    :param pre:
    :type pre: str
    :param post:
    :type post: str
    :rtype: None
    """
    event.current_buffer.cursor_position += (
        event.current_buffer.document.get_start_of_line_position(
            after_whitespace=True
        )
    )
    event.cli.current_buffer.insert_text(pre)
    event.current_buffer.cursor_position += (
        event.current_buffer.document.get_end_of_line_position()
    )
    event.cli.current_buffer.insert_text(post)
    event.current_buffer.validate_and_handle()



r"""Prompt Toolkit
==================
"""

import os
from dataclasses import dataclass

from ptpython.repl import PythonRepl
from wcwidth import wcswidth

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.layout.containers import (
    Float,
    FloatContainer,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import Frame

from .. import (
    clearComposition,
    commitComposition,
    createSession,
    getCommit,
    getContext,
    init,
    processKey,
)
from ..__main__ import Traits
from ..draw_ui import UI, draw_ui
from ..parse_key import (
    ALT_SHIFT_CR,
    CONTROL_ALT_CR,
    CONTROL_ALT_SHIFT_CR,
    CONTROL_CR,
    CONTROL_SHIFT_CR,
    SHIFT_CR,
    parse_keys,
)


@dataclass
class Rime:
    r"""Rime."""

    repl: PythonRepl
    session_id: int = 0
    is_enabled: bool = False
    remember_rime: bool = False
    preedit: str = ""
    window: Window = None  # type: ignore
    layout: Layout | None = None
    traits: Traits = None  # type: ignore
    ui: UI = None  # type: ignore
    keys_set: set[tuple[str, ...]] = None  # type: ignore

    def __post_init__(self) -> None:
        r"""Post init.

        :rtype: None
        """
        if self.traits is None:
            self.traits = Traits()
        if self.ui is None:
            self.ui = UI()
        if self.keys_set is None:
            self.keys_set = {
                ("s-tab",),
                ("s-escape",),
                ("escape", "backspace"),
                tuple(SHIFT_CR),
                tuple(ALT_SHIFT_CR),
                tuple(CONTROL_CR),
                tuple(CONTROL_SHIFT_CR),
                tuple(CONTROL_ALT_CR),
                tuple(CONTROL_ALT_SHIFT_CR),
            }
            for order in range(ord(" "), ord("~") + 1):
                self.keys_set |= {(chr(order),)}
            for number in range(1, 24):
                self.keys_set |= {(f"f{number}",)}
            for keyname in {
                "insert",
                "delete",
                "up",
                "down",
                "left",
                "right",
                "home",
                "end",
                "pageup",
                "pagedown",
            }:
                self.keys_set |= {
                    (keyname,),
                    ("c-" + keyname,),
                    ("s-" + keyname,),
                    ("c-s-" + keyname,),
                    ("escape", keyname),
                    ("escape", "c-" + keyname),
                    ("escape", "s-" + keyname),
                    ("escape", "c-s-" + keyname),
                }
            for order in range(ord("@"), ord("[")):
                key = "c-" + chr(order).lower()
                self.keys_set |= {(key,), ("escape", key)}
            self.keys_set |= {("escape", "escape")}
            for order in range(ord("[") + 1, ord("_")):
                key = "c-" + chr(order)
                self.keys_set |= {(key,), ("escape", key)}
            for order in range(ord(" "), ord("~") + 1):
                self.keys_set |= {("escape", chr(order))}

        for keys in self.keys_set:
            keys = list(keys)

            @self.repl.add_key_binding(*keys, filter=self.mode(keys))  # type: ignore
            def _(event: KeyPressEvent, keys: list[str] = keys) -> None:
                r""".

                :param event:
                :type event: KeyPressEvent
                :param keys:
                :type keys: list[str]
                :rtype: None
                """
                self.key_binding(event, keys)

    def key_binding(self, event: KeyPressEvent, keys: list[str]) -> None:
        r"""Key binding.

        :param event:
        :type event: KeyPressEvent
        :param keys:
        :type keys: list[str]
        :rtype: None
        """
        text, lines, col = self.draw(keys)
        self.preedit = lines[0].strip(" " + self.ui.cursor)
        ui_text = "\n".join(lines)
        event.cli.current_buffer.insert_text(text)
        self.window.height = len(lines)
        if len(lines):
            self.window.width = max(wcswidth(line) for line in lines)
        self.window.content.buffer.text = ui_text  # type: ignore
        left, top = self.calculate()
        left += col
        self.repl.app.layout.container.floats[0].left = left  # type: ignore
        self.repl.app.layout.container.floats[0].top = top  # type: ignore

    def get_commit_text(self) -> str:
        r"""Get commit text.

        :rtype: str
        """
        if commitComposition(self.session_id):
            return getCommit(self.session_id).text
        return ""

    def draw(self, keys: list[str]) -> tuple[str, list[str], int]:
        r"""Draw.

        :param keys:
        :type keys: list[str]
        :rtype: tuple[str, list[str], int]
        """
        if not processKey(self.session_id, *parse_keys(keys)):
            if len(keys) == 1 == len(keys[0]):
                return keys[0], [self.ui.cursor], 0
            return "", [self.ui.cursor], 0
        context = getContext(self.session_id)
        if context.menu.num_candidates == 0:
            return self.get_commit_text(), [self.ui.cursor], 0
        lines, col = draw_ui(context, self.ui)
        return "", lines, col

    def mode(self, keys: list[str]) -> Condition:
        r"""Mode.

        :param keys:
        :type keys: list[str]
        :rtype: Condition
        """

        @Condition
        def _(keys: list[str] = keys) -> bool:
            r""".

            :param keys:
            :type keys: list[str]
            :rtype: bool
            """
            if len(keys) == 1 == len(keys[0]):
                return self.is_enabled
            elif len(keys) == 1 or len(keys) > 1 and keys[0] == "escape":
                return self.preedit != ""
            else:
                raise NotImplementedError

        return _

    def filter(self, condition: Condition | None = None) -> Condition:
        r"""Filter.

        :param condition:
        :type condition: Condition | None
        :rtype: Condition
        """

        @Condition
        def _(condition: Condition | None = condition) -> bool:
            r""".

            :param condition:
            :type condition: Condition | None
            :rtype: bool
            """
            if condition is None:
                return self.preedit == ""
            return self.preedit == "" and condition()

        return _

    def swap_layout(self):
        r"""Swap layout."""
        (self.layout, self.repl.app.layout) = (  # type: ignore
            self.repl.app.layout,
            self.layout,
        )

    def disable(self) -> None:
        r"""Disable.

        :rtype: None
        """
        self.swap_layout()
        self.is_enabled = False
        self.preedit = ""
        clearComposition(self.session_id)

    def calculate(self) -> tuple[int, int]:
        r"""Calculate.

        :rtype: tuple[int, int]
        """
        left = 0
        for _, text in self.repl.all_prompt_styles[  # type: ignore
            self.repl.prompt_style
        ].in_prompt():
            left += wcswidth(text)
        top = 0
        if self.repl.app.layout.current_buffer:
            lines = self.repl.app.layout.current_buffer.text[
                : self.repl.app.layout.current_buffer.cursor_position
            ].split("\n")
            top += len(lines)
            left += wcswidth(lines[-1])
        return left, top

    def init_(self) -> None:
        r"""Init.

        :rtype: None
        """
        if self.session_id == 0:
            os.makedirs(self.traits.log_dir, exist_ok=True)
            init(**vars(self.traits))
            self.session_id = createSession()

    def enable(self) -> None:
        r"""Enable.

        :rtype: None
        """
        self.init_()
        left, top = self.calculate()
        self.window = Window(
            BufferControl(buffer=Buffer()),
            width=wcswidth(self.ui.cursor),
            height=1,
        )
        self.window.content.buffer.text = self.ui.cursor  # type: ignore
        self.layout = Layout(
            FloatContainer(  # type: ignore
                self.repl.app.layout.container,
                [
                    Float(
                        Frame(
                            self.window,
                        ),
                        left=left,
                        top=top,
                    )
                ],
            )
        )
        self.swap_layout()
        self.is_enabled = True

    def conditional_disable(self) -> None:
        r"""Conditional disable.

        :rtype: None
        """
        if self.is_enabled:
            self.remember_rime = True
            self.disable()

    def conditional_enable(self) -> None:
        r"""Conditional enable.

        :rtype: None
        """
        if self.remember_rime:
            self.enable()

    def toggle(self) -> None:
        r"""Toggle.

        :rtype: None
        """
        if self.is_enabled:
            self.disable()
        else:
            self.enable()



"""Prompt style
===============

Generate prompt style from ``sys.ps1``.

Experimental!
"""

# import sys
from typing import TYPE_CHECKING

from prompt_toolkit.formatted_text import AnyFormattedText
from ptpython.prompt_style import IPythonPrompt

# from .ansi import ansi2prompt_toolkit

if TYPE_CHECKING:
    from ptpython.python_input import PythonInput


class PythonPrompt(IPythonPrompt):
    """A prompt resembling the IPython prompt."""

    def __init__(self, python_input: "PythonInput") -> None:
        """Init.

        :param python_input:
        :type python_input: "PythonInput"
        :rtype: None
        """
        super().__init__(python_input)
        # self.formats, self.styles = ansi2prompt_toolkit(str(sys.ps1))

    def in_prompt(self) -> AnyFormattedText:
        """In prompt.

        :rtype: AnyFormattedText
        """
        # self.formats, self.styles = ansi2prompt_toolkit(str(sys.ps1))
        return [
            ("class:in", "In ["),
            (
                "class:in.number",
                f"{self.python_input.current_statement_index}",
            ),
            ("class:in", "]: "),
        ]
